@@ -1,3 +1,5 @@
+#![allow(clippy::borrow_deref_ref)]
+
 use std::{
     convert::Infallible,
     fmt::Display,
@@ -6,17 +8,17 @@ use std::{
 };
 
 use crate::common::{
-    self, DConfig, DHeader, DImage, DImageD, DSeriesEnd, DetectorConfig, FrameData, FrameSender,
-    PixelType, TriggerMode,
+    self, setup_monitor, DConfig, DHeader, DImage, DImageD, DSeriesEnd, DetectorConfig, FrameData,
+    FrameSender, PixelType, TriggerMode,
 };
 
 use bincode::serialize;
 use crossbeam_channel::{unbounded, Receiver, RecvTimeoutError, SendError, Sender, TryRecvError};
-use log::{info, debug};
+use log::{debug, info};
 use pyo3::{
-    exceptions,
+    create_exception, exceptions,
     prelude::*,
-    types::{PyBytes, PyType}, create_exception,
+    types::{PyBytes, PyType},
 };
 use zmq::{Message, Socket};
 
@@ -26,15 +28,28 @@ fn libertem_dectris(py: Python, m: &PyModule) -> PyResult<()> {
     // the GIL
     // pyo3_log::init();
 
-    m.add_class::<Frame>().unwrap();
-    m.add_class::<FrameIterator>().unwrap();
-    m.add_class::<FrameStack>().unwrap();
-    m.add_class::<FrameChunkedIterator>().unwrap();
-    m.add_class::<PixelType>().unwrap();
-    m.add_class::<DectrisSim>().unwrap();
-    m.add_class::<DetectorConfig>().unwrap();
-    m.add_class::<TriggerMode>().unwrap();
+    m.add_class::<Frame>()?;
+    m.add_class::<FrameIterator>()?;
+    m.add_class::<FrameStack>()?;
+    m.add_class::<FrameChunkedIterator>()?;
+    m.add_class::<PixelType>()?;
+    m.add_class::<DectrisSim>()?;
+    m.add_class::<DetectorConfig>()?;
+    m.add_class::<TriggerMode>()?;
     m.add("TimeoutError", py.get_type::<TimeoutError>())?;
+
+    register_header_module(py, m)?;
+    Ok(())
+}
+
+fn register_header_module(py: Python<'_>, parent_module: &PyModule) -> PyResult<()> {
+    let headers_module = PyModule::new(py, "headers")?;
+    headers_module.add_class::<DHeader>()?;
+    headers_module.add_class::<DImage>()?;
+    headers_module.add_class::<DImageD>()?;
+    headers_module.add_class::<DConfig>()?;
+    headers_module.add_class::<DSeriesEnd>()?;
+    parent_module.add_submodule(headers_module)?;
     Ok(())
 }
 
@@ -50,25 +65,33 @@ impl Frame {
             frame: frame.clone(),
         }
     }
-
-    fn with_data(frame: FrameData) -> Self {
-        Frame { frame }
-    }
 }
 
 #[pymethods]
 impl Frame {
+    #[new]
+    fn new(data: &PyBytes, dimage: &DImage, dimaged: &DImageD, dconfig: &DConfig) -> Self {
+        let frame_data: FrameData = FrameData {
+            dimage: dimage.clone(),
+            dimaged: dimaged.clone(),
+            image_data: data.as_bytes().into(),
+            dconfig: dconfig.clone(),
+        };
+
+        Frame { frame: frame_data }
+    }
+
     fn __repr__(&self) -> String {
         let frame = &self.frame;
         let series = frame.dimage.series;
         let shape = &frame.dimaged.shape;
         let idx = frame.dimage.frame;
-        return format!("<Frame idx={idx} series={series} shape={shape:?}>");
+        format!("<Frame idx={idx} series={series} shape={shape:?}>")
     }
 
     fn get_image_data(slf: PyRef<Self>, py: Python) -> Py<PyBytes> {
-        let bytes: &PyBytes = PyBytes::new(py, &slf.frame.image_data.as_slice());
-        return bytes.into();
+        let bytes: &PyBytes = PyBytes::new(py, slf.frame.image_data.as_slice());
+        bytes.into()
     }
 
     fn get_series_id(slf: PyRef<Self>) -> u64 {
@@ -84,35 +107,47 @@ impl Frame {
     }
 
     fn get_pixel_type(slf: PyRef<Self>) -> String {
-        return match &slf.frame.dimaged.type_ {
+        match &slf.frame.dimaged.type_ {
             PixelType::Uint8 => "uint8".to_string(),
             PixelType::Uint16 => "uint16".to_string(),
             PixelType::Uint32 => "uint32".to_string(),
-        };
+        }
     }
 
     fn get_encoding(slf: PyRef<Self>) -> String {
         slf.frame.dimaged.encoding.clone()
     }
 
+    /// return endianess in numpy notation
+    fn get_endianess(slf: PyRef<Self>) -> String {
+        let last_char = slf.frame.dimaged.encoding.chars().last();
+        last_char.expect("encoding should be non-empty").into()
+    }
+
     fn get_shape(slf: PyRef<Self>) -> Vec<u64> {
         slf.frame.dimaged.shape.clone()
     }
 }
 
+impl From<Frame> for FrameData {
+    fn from(frame: Frame) -> Self {
+        frame.frame
+    }
+}
+
 pub enum ControlMsg {
     StopThread,
     StartAcquisition { series: u64 },
 }
 
-#[derive(PartialEq)]
+#[derive(PartialEq, Eq)]
 pub enum ResultMsg {
     Error { msg: String }, // generic error response, might need to specialize later
     Frame { frame: FrameData },
     End,
 }
 
-#[derive(PartialEq)]
+#[derive(PartialEq, Eq)]
 pub enum ReceiverStatus {
     Idle,
     Running,
@@ -137,8 +172,8 @@ fn recv_part(
             Err(zmq::Error::EAGAIN) => {
                 check_for_control(control_channel)?;
                 continue;
-            },
-            Err(err) => AcquisitionError::ZmqError { err }
+            }
+            Err(err) => AcquisitionError::ZmqError { err },
         };
     }
     Ok(())
@@ -184,24 +219,31 @@ enum AcquisitionError {
 
 impl Display for AcquisitionError {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "acquisition has stopped")
+        match self {
+            AcquisitionError::ZmqError { err } => {
+                write!(f, "zmq error {err}")
+            }
+            AcquisitionError::Cancelled => {
+                write!(f, "acquisition cancelled")
+            }
+            AcquisitionError::SeriesMismatch => {
+                write!(f, "series mismatch")
+            }
+            AcquisitionError::Disconnected => {
+                write!(f, "other end has disconnected")
+            }
+        }
     }
 }
 
-fn check_for_control(
-    control_channel: &Receiver<ControlMsg>,
-) -> Result<(), AcquisitionError> {
+fn check_for_control(control_channel: &Receiver<ControlMsg>) -> Result<(), AcquisitionError> {
     match control_channel.try_recv() {
         Ok(ControlMsg::StartAcquisition { series: _ }) => {
             panic!("received StartAcquisition while an acquisition was already running");
         }
-        Ok(ControlMsg::StopThread) => {
-            return Err(AcquisitionError::Cancelled);
-        }
-        Err(TryRecvError::Disconnected) => {
-            return Err(AcquisitionError::Cancelled);
-        }
-        Err(TryRecvError::Empty) => Ok(())
+        Ok(ControlMsg::StopThread) => Err(AcquisitionError::Cancelled),
+        Err(TryRecvError::Disconnected) => Err(AcquisitionError::Cancelled),
+        Err(TryRecvError::Empty) => Ok(()),
     }
 }
 
@@ -217,10 +259,10 @@ fn acquisition(
     loop {
         if last_control_check.elapsed() > Duration::from_millis(300) {
             last_control_check = Instant::now();
-            check_for_control(&to_thread_r)?;
+            check_for_control(to_thread_r)?;
         }
 
-        let frame = recv_frame(&socket, &to_thread_r)?;
+        let frame = recv_frame(socket, to_thread_r)?;
 
         if frame.dimage.series != series {
             return Err(AcquisitionError::SeriesMismatch);
@@ -255,55 +297,58 @@ fn acquisition(
     }
 }
 
-
 /// convert `AcquisitionError`s to messages on `from_threads_s`
-fn background_thread_wrap(to_thread_r: &Receiver<ControlMsg>, from_thread_s: &Sender<ResultMsg>) {
-    match background_thread(to_thread_r, from_thread_s) {
-        Ok(()) => {},
-        Err(err) => {
-            from_thread_s
-                .send(ResultMsg::Error {
-                    msg: err.to_string()
-                })
-                .unwrap();
-            return;
-        }
+fn background_thread_wrap(
+    to_thread_r: &Receiver<ControlMsg>,
+    from_thread_s: &Sender<ResultMsg>,
+    uri: String,
+) {
+    if let Err(err) = background_thread(to_thread_r, from_thread_s, uri) {
+        from_thread_s
+            .send(ResultMsg::Error {
+                msg: err.to_string(),
+            })
+            .unwrap();
     }
 }
 
 fn background_thread(
-    to_thread_r: &Receiver<ControlMsg>, from_thread_s: &Sender<ResultMsg>
+    to_thread_r: &Receiver<ControlMsg>,
+    from_thread_s: &Sender<ResultMsg>,
+    uri: String,
 ) -> Result<(), AcquisitionError> {
     let ctx = zmq::Context::new();
     let socket = ctx.socket(zmq::PULL).unwrap();
     socket.set_rcvtimeo(1000).unwrap();
-    socket.connect("tcp://127.0.0.1:9999").unwrap();
+    socket.connect(&uri).unwrap();
     socket.set_rcvhwm(4 * 256).unwrap();
 
+    setup_monitor(ctx, "DectrisReceiver".to_string(), &socket);
+
     loop {
         // control: main threads tells us to quit
         let control = to_thread_r.recv_timeout(Duration::from_millis(100));
         match control {
             Ok(ControlMsg::StartAcquisition { series }) => {
                 let mut msg: Message = Message::new();
-                recv_part(&mut msg, &socket, &to_thread_r)?;
+                recv_part(&mut msg, &socket, to_thread_r)?;
                 // panic in case of any other header type:
                 let dheader: DHeader = serde_json::from_str(msg.as_str().unwrap()).unwrap();
                 debug!("dheader: {dheader:?}");
 
                 // second message: the header itself
-                recv_part(&mut msg, &socket, &to_thread_r)?;
+                recv_part(&mut msg, &socket, to_thread_r)?;
                 let detector_config: DetectorConfig =
                     serde_json::from_str(msg.as_str().unwrap()).unwrap();
 
                 match acquisition(detector_config, to_thread_r, from_thread_s, &socket, series) {
-                    Ok(_) => {},
+                    Ok(_) => {}
                     Err(AcquisitionError::Disconnected | AcquisitionError::Cancelled) => {
                         return Ok(());
-                    },
+                    }
                     e => {
                         return e;
-                    },
+                    }
                 }
             }
             Ok(ControlMsg::StopThread) => {
@@ -330,17 +375,20 @@ impl Display for ReceiverError {
 }
 
 impl DectrisReceiver {
-    pub fn new() -> Self {
+    pub fn new(uri: &str) -> Self {
         let (to_thread_s, to_thread_r) = unbounded();
         let (from_thread_s, from_thread_r) = unbounded();
 
         let builder = std::thread::Builder::new();
+        let uri = uri.to_string();
 
         DectrisReceiver {
             bg_thread: Some(
                 builder
                     .name("bg_thread".to_string())
-                    .spawn(move || background_thread_wrap(&to_thread_r, &from_thread_s))
+                    .spawn(move || {
+                        background_thread_wrap(&to_thread_r, &from_thread_s, uri.to_string())
+                    })
                     .expect("failed to start background thread"),
             ),
             from_thread: from_thread_r,
@@ -349,7 +397,7 @@ impl DectrisReceiver {
         }
     }
 
-    pub fn next(&mut self) -> ResultMsg {
+    pub fn recv(&mut self) -> ResultMsg {
         let result_msg = self
             .from_thread
             .recv()
@@ -357,7 +405,7 @@ impl DectrisReceiver {
         if result_msg == ResultMsg::End {
             self.status = ReceiverStatus::Idle;
         }
-        return result_msg;
+        result_msg
     }
 
     pub fn next_timeout(&mut self, timeout: Duration) -> Option<ResultMsg> {
@@ -368,13 +416,13 @@ impl DectrisReceiver {
                 if result == ResultMsg::End {
                     self.status = ReceiverStatus::Idle;
                 }
-                return Some(result);
+                Some(result)
             }
             Err(e) => match e {
                 RecvTimeoutError::Disconnected => {
                     panic!("background thread should be running")
                 }
-                RecvTimeoutError::Timeout => return None,
+                RecvTimeoutError::Timeout => None,
             },
         }
     }
@@ -405,6 +453,12 @@ impl DectrisReceiver {
     }
 }
 
+impl Default for DectrisReceiver {
+    fn default() -> Self {
+        Self::new("tcp://127.0.0.1:9999")
+    }
+}
+
 #[pyclass]
 pub struct FrameIterator {
     receiver: DectrisReceiver,
@@ -413,16 +467,16 @@ pub struct FrameIterator {
 #[pymethods]
 impl FrameIterator {
     #[new]
-    fn new() -> Self {
+    fn new(uri: &str) -> Self {
         FrameIterator {
-            receiver: DectrisReceiver::new(),
+            receiver: DectrisReceiver::new(uri),
         }
     }
 
     fn start(mut slf: PyRefMut<Self>, series: u64) -> PyResult<()> {
         slf.receiver
             .start(series)
-            .map_err(|err| return exceptions::PyRuntimeError::new_err(err.msg))
+            .map_err(|err| exceptions::PyRuntimeError::new_err(err.msg))
     }
 
     fn close(mut slf: PyRefMut<Self>) {
@@ -440,7 +494,9 @@ impl FrameIterator {
                     return Err(exceptions::PyRuntimeError::new_err(msg));
                 }
                 Some(ResultMsg::End) => return Ok(None),
-                Some(ResultMsg::Frame { frame }) => return Ok(Some(Frame::with_data_cloned(&frame))),
+                Some(ResultMsg::Frame { frame }) => {
+                    return Ok(Some(Frame::with_data_cloned(&frame)))
+                }
                 None => {
                     py.check_signals()?;
                     py.allow_threads(|| {
@@ -493,15 +549,24 @@ impl FrameStack {
         FrameStack::empty()
     }
 
+    /// create a frame stack from a list of frames
+    /// NOTE: this probably doesn't perform well, and is only meant for testing
+    #[classmethod]
+    fn from_frame_list(_cls: &PyType, frames: Vec<Frame>) -> Self {
+        FrameStack {
+            frames: frames.into_iter().map(|f| f.into()).collect(),
+        }
+    }
+
     #[classmethod]
     fn deserialize(_cls: &PyType, serialized: &PyBytes) -> Self {
         let data = serialized.as_bytes();
-        FrameStack::with_data(bincode::deserialize(&data).unwrap())
+        FrameStack::with_data(bincode::deserialize(data).unwrap())
     }
 
     fn serialize(slf: PyRef<Self>, py: Python) -> PyResult<Py<PyBytes>> {
         let bytes: &PyBytes = PyBytes::new(py, serialize(&slf.frames).unwrap().as_slice());
-        return Ok(bytes.into());
+        Ok(bytes.into())
     }
 
     fn __len__(slf: PyRef<Self>) -> usize {
@@ -525,16 +590,16 @@ struct FrameChunkedIterator {
 #[pymethods]
 impl FrameChunkedIterator {
     #[new]
-    fn new() -> Self {
+    fn new(uri: &str) -> Self {
         FrameChunkedIterator {
-            receiver: DectrisReceiver::new(),
+            receiver: DectrisReceiver::new(uri),
         }
     }
 
     fn start(mut slf: PyRefMut<Self>, series: u64) -> PyResult<()> {
         slf.receiver
             .start(series)
-            .map_err(|err| return exceptions::PyRuntimeError::new_err(err.msg))
+            .map_err(|err| exceptions::PyRuntimeError::new_err(err.msg))
     }
 
     fn close(mut slf: PyRefMut<Self>) {
@@ -542,7 +607,7 @@ impl FrameChunkedIterator {
     }
 
     fn is_running(slf: PyRef<Self>) -> bool {
-        return slf.receiver.status == ReceiverStatus::Running;
+        slf.receiver.status == ReceiverStatus::Running
     }
 
     fn get_next_stack(
@@ -568,7 +633,7 @@ impl FrameChunkedIterator {
             let recv_result = py.allow_threads(|| {
                 let next: Result<Option<ResultMsg>, Infallible> =
                     Ok(recv.next_timeout(Duration::from_millis(100)));
-                return next;
+                next
             })?;
 
             match recv_result {
@@ -602,25 +667,43 @@ create_exception!(
 #[pyclass]
 struct DectrisSim {
     frame_sender: FrameSender,
-    dwelltime: Option<u64>,  // in µseconds
+    dwelltime: Option<u64>, // in µseconds
 }
 
 #[pymethods]
 impl DectrisSim {
     #[new]
-    fn new(uri: &str, filename: &str, dwelltime: Option<u64>) -> Self {
-        return DectrisSim {
-            frame_sender: FrameSender::new(uri, filename),
+    fn new(uri: &str, filename: &str, dwelltime: Option<u64>, random_port: bool) -> Self {
+        DectrisSim {
+            frame_sender: FrameSender::new(uri, filename, random_port),
             dwelltime,
-        };
+        }
+    }
+
+    fn get_uri(slf: PyRef<Self>) -> String {
+        slf.frame_sender.get_uri().to_string()
     }
 
     fn get_detector_config(slf: PyRef<Self>) -> DetectorConfig {
         slf.frame_sender.get_detector_config().clone()
     }
 
-    fn send_headers(mut slf: PyRefMut<Self>) {
-        slf.frame_sender.send_headers();
+    fn send_headers(mut slf: PyRefMut<Self>, py: Python) -> PyResult<()> {
+        let sender = &mut slf.frame_sender;
+        py.allow_threads(|| {
+            if let Err(e) = sender.send_headers(|| {
+                let gil = Python::acquire_gil();
+                if let Err(e) = gil.python().check_signals() {
+                    eprintln!("got python error {e:?}, breaking");
+                    return None;
+                }
+                Some(())
+            }) {
+                let msg = format!("failed to send headers: {e:?}");
+                return Err(exceptions::PyRuntimeError::new_err(msg));
+            }
+            Ok(())
+        })
     }
 
     /// send `nframes`, if given, or all frames in the acquisition, from the
@@ -634,29 +717,25 @@ impl DectrisSim {
             Some(n) => n,
         };
 
+        let dwelltime = &slf.dwelltime.clone();
+        let sender = &mut slf.frame_sender;
+
         for frame_idx in 0..effective_nframes {
-            // can't `allow_threads` because the zmq socket is not Send
-            //py.allow_threads(|| {
-                match slf.frame_sender.send_frame() {
-                    Err(common::SendError::Timeout) => {
-                        return Err(TimeoutError::new_err(
-                            "timeout while sending frames".to_string(),
-                        ));
-                    }
-                    Err(_) => {
-                        return Err(exceptions::PyRuntimeError::new_err(
-                            "error while sending frames".to_string(),
-                        ));
-                    }
-                    Ok(_) => {} // Ok(())
-                }
-            //})?;
+            py.allow_threads(|| match sender.send_frame() {
+                Err(common::SendError::Timeout) => Err(TimeoutError::new_err(
+                    "timeout while sending frames".to_string(),
+                )),
+                Err(_) => Err(exceptions::PyRuntimeError::new_err(
+                    "error while sending frames".to_string(),
+                )),
+                Ok(_) => Ok(()),
+            })?;
 
             // dwelltime
             // FIXME: for continuous mode, u64 might not be enough for elapsed time,
             // so maybe it's better to carry around a "budget" that can be negative
             // if a frame hasn't been sent out in time etc.
-            if let Some(dt) = slf.dwelltime {
+            if let Some(dt) = dwelltime {
                 let elapsed_us = start_time.elapsed().as_micros() as u64;
                 let target_time_us = (frame_idx + 1) * dt;
                 if elapsed_us < target_time_us {
@@ -671,7 +750,7 @@ impl DectrisSim {
                 py.check_signals()?;
 
                 // also drop GIL once in a while
-                py.allow_threads(|| { 
+                py.allow_threads(|| {
                     spin_sleep::sleep(Duration::from_micros(5));
                 });
             }
